@@ -1,12 +1,14 @@
+#[cfg(not(feature = "std"))]
+use alloc::boxed::Box;
+#[cfg(not(feature = "std"))]
 use alloc::vec;
+#[cfg(not(feature = "std"))]
+use alloc::vec::Vec;
 use core::borrow::Borrow;
-use core::convert::TryInto;
 use core::hash::{Hash, Hasher};
 use core::marker::PhantomData;
 use core::mem;
 
-use chrono::{Datelike, TimeZone, Timelike};
-
 use crate::writer::Writer;
 use crate::{
     parse, parse_single, BitString, ObjectIdentifier, OwnedBitString, ParseError, ParseErrorKind,
@@ -44,9 +46,17 @@ impl<'a, T: SimpleAsn1Readable<'a>> Asn1Readable<'a> for T {
     }
 }
 
+impl<'a, T: SimpleAsn1Readable<'a>> SimpleAsn1Readable<'a> for Box<T> {
+    const TAG: Tag = T::TAG;
+
+    fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
+        Ok(Box::new(T::parse_data(data)?))
+    }
+}
+
 /// Any type that can be written as DER ASN.1.
 pub trait Asn1Writable: Sized {
-    fn write(&self, dest: &mut Writer) -> WriteResult;
+    fn write(&self, dest: &mut Writer<'_>) -> WriteResult;
 }
 
 // Types with a fixed-tag that can be written as DER ASN.1.
@@ -56,9 +66,18 @@ pub trait SimpleAsn1Writable: Sized {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult;
 }
 
+pub trait Asn1DefinedByReadable<'a, T: Asn1Readable<'a>>: Sized {
+    fn parse(item: T, parser: &mut Parser<'a>) -> ParseResult<Self>;
+}
+
+pub trait Asn1DefinedByWritable<T: Asn1Writable>: Sized {
+    fn item(&self) -> &T;
+    fn write(&self, dest: &mut Writer<'_>) -> WriteResult;
+}
+
 impl<T: SimpleAsn1Writable> Asn1Writable for T {
     #[inline]
-    fn write(&self, w: &mut Writer) -> WriteResult {
+    fn write(&self, w: &mut Writer<'_>) -> WriteResult {
         w.write_tlv(Self::TAG, move |dest| self.write_data(dest))
     }
 }
@@ -70,6 +89,13 @@ impl<T: SimpleAsn1Writable> SimpleAsn1Writable for &T {
     }
 }
 
+impl<T: SimpleAsn1Writable> SimpleAsn1Writable for Box<T> {
+    const TAG: Tag = T::TAG;
+    fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
+        T::write_data(self, dest)
+    }
+}
+
 /// A TLV (type, length, value) represented as the tag and bytes content.
 /// Generally used for parsing ASN.1 `ANY` values.
 #[derive(Debug, PartialEq, Hash, Clone, Copy, Eq)]
@@ -114,7 +140,7 @@ impl<'a> Asn1Readable<'a> for Tlv<'a> {
 }
 impl<'a> Asn1Writable for Tlv<'a> {
     #[inline]
-    fn write(&self, w: &mut Writer) -> WriteResult {
+    fn write(&self, w: &mut Writer<'_>) -> WriteResult {
         w.write_tlv(self.tag, move |dest| dest.push_slice(self.data))
     }
 }
@@ -179,6 +205,54 @@ impl<'a> SimpleAsn1Writable for &'a [u8] {
     }
 }
 
+impl<const N: usize> SimpleAsn1Readable<'_> for [u8; N] {
+    const TAG: Tag = Tag::primitive(0x04);
+    fn parse_data(data: &[u8]) -> ParseResult<[u8; N]> {
+        data.try_into()
+            .map_err(|_| ParseError::new(ParseErrorKind::InvalidValue))
+    }
+}
+
+impl<const N: usize> SimpleAsn1Writable for [u8; N] {
+    const TAG: Tag = Tag::primitive(0x04);
+    fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
+        dest.push_slice(self)
+    }
+}
+
+/// Represents values that are encoded as an `OCTET STRING` containing an
+/// encoded TLV, of type `T`.
+#[derive(PartialEq, Eq, Debug)]
+pub struct OctetStringEncoded<T>(T);
+
+impl<T> OctetStringEncoded<T> {
+    pub fn new(v: T) -> OctetStringEncoded<T> {
+        OctetStringEncoded(v)
+    }
+
+    pub fn get(&self) -> &T {
+        &self.0
+    }
+
+    pub fn into_inner(self) -> T {
+        self.0
+    }
+}
+
+impl<'a, T: Asn1Readable<'a>> SimpleAsn1Readable<'a> for OctetStringEncoded<T> {
+    const TAG: Tag = Tag::primitive(0x04);
+    fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
+        Ok(OctetStringEncoded::new(parse_single(data)?))
+    }
+}
+
+impl<T: Asn1Writable> SimpleAsn1Writable for OctetStringEncoded<T> {
+    const TAG: Tag = Tag::primitive(0x04);
+    fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
+        self.0.write(&mut Writer::new(dest))
+    }
+}
+
 /// Type for use with `Parser.read_element` and `Writer.write_element` for
 /// handling ASN.1 `PrintableString`.  A `PrintableString` contains an `&str`
 /// with only valid characers.
@@ -265,7 +339,7 @@ impl<'a> IA5String<'a> {
         }
     }
 
-    fn new_from_bytes(s: &'a [u8]) -> Option<IA5String> {
+    fn new_from_bytes(s: &'a [u8]) -> Option<IA5String<'a>> {
         if IA5String::verify(s) {
             // TODO: This value is always valid utf-8 because we just verified
             // the contents, but I don't want to call an unsafe function, so we
@@ -309,7 +383,7 @@ impl<'a> Utf8String<'a> {
         Utf8String(s)
     }
 
-    fn new_from_bytes(s: &'a [u8]) -> Option<Utf8String> {
+    fn new_from_bytes(s: &'a [u8]) -> Option<Utf8String<'a>> {
         Some(Utf8String(core::str::from_utf8(s).ok()?))
     }
 
@@ -347,7 +421,7 @@ impl<'a> VisibleString<'a> {
         }
     }
 
-    fn new_from_bytes(s: &'a [u8]) -> Option<VisibleString> {
+    fn new_from_bytes(s: &'a [u8]) -> Option<VisibleString<'a>> {
         if VisibleString::verify(s) {
             // TODO: This value is always valid utf-8 because we just verified
             // the contents, but I don't want to call an unsafe function, so we
@@ -555,6 +629,8 @@ macro_rules! impl_asn1_element_for_int {
 
 impl_asn1_element_for_int!(i8; true);
 impl_asn1_element_for_int!(u8; false);
+impl_asn1_element_for_int!(i16; true);
+impl_asn1_element_for_int!(u16; false);
 impl_asn1_element_for_int!(i32; true);
 impl_asn1_element_for_int!(u32; false);
 impl_asn1_element_for_int!(i64; true);
@@ -597,6 +673,44 @@ impl<'a> SimpleAsn1Writable for BigUint<'a> {
     }
 }
 
+/// Arbitrary sized unsigned integer which owns its data. Contents may be
+/// accessed as `&[u8]` of big-endian data. Its contents always match the DER
+/// encoding of a value (i.e. they are minimal)
+#[derive(PartialEq, Clone, Debug, Hash, Eq)]
+pub struct OwnedBigUint {
+    data: Vec<u8>,
+}
+
+impl OwnedBigUint {
+    /// Create a new `OwnedBigUint` from already encoded data. `data` must be
+    /// encoded as required by DER: minimally and if the high bit would be set
+    /// in the first octet, a leading `\x00` should be prepended (to
+    /// disambiguate from negative values).
+    pub fn new(data: Vec<u8>) -> Option<Self> {
+        validate_integer(&data, false).ok()?;
+        Some(OwnedBigUint { data })
+    }
+
+    /// Returns the contents of the integer as big-endian bytes.
+    pub fn as_bytes(&self) -> &[u8] {
+        &self.data
+    }
+}
+
+impl SimpleAsn1Readable<'_> for OwnedBigUint {
+    const TAG: Tag = Tag::primitive(0x02);
+    fn parse_data(data: &[u8]) -> ParseResult<Self> {
+        OwnedBigUint::new(data.to_vec())
+            .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
+    }
+}
+impl SimpleAsn1Writable for OwnedBigUint {
+    const TAG: Tag = Tag::primitive(0x02);
+    fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
+        dest.push_slice(self.as_bytes())
+    }
+}
+
 /// Arbitrary sized signed integer. Contents may be accessed as `&[u8]` of
 /// big-endian data. Its contents always match the DER encoding of a value
 /// (i.e. they are minimal)
@@ -619,6 +733,11 @@ impl<'a> BigInt<'a> {
     pub fn as_bytes(&self) -> &'a [u8] {
         self.data
     }
+
+    /// Returns a boolean indicating whether the integer is negative.
+    pub fn is_negative(&self) -> bool {
+        self.data[0] & 0x80 == 0x80
+    }
 }
 
 impl<'a> SimpleAsn1Readable<'a> for BigInt<'a> {
@@ -634,6 +753,48 @@ impl<'a> SimpleAsn1Writable for BigInt<'a> {
     }
 }
 
+/// Arbitrary sized signed integer which owns its contents. Contents may be
+/// accessed as `&[u8]` of big-endian data. Its contents always match the DER
+/// encoding of a value (i.e. they are minimal)
+#[derive(PartialEq, Clone, Debug, Hash, Eq)]
+pub struct OwnedBigInt {
+    data: Vec<u8>,
+}
+
+impl OwnedBigInt {
+    /// Create a new `OwnedBigInt` from already encoded data. `data` must be
+    /// encoded as required by DER: minimally and if the high bit would be set
+    /// in the first octet, a leading `\x00` should be prepended (to
+    /// disambiguate from negative values).
+    pub fn new(data: Vec<u8>) -> Option<Self> {
+        validate_integer(&data, true).ok()?;
+        Some(OwnedBigInt { data })
+    }
+
+    /// Returns the contents of the integer as big-endian bytes.
+    pub fn as_bytes(&self) -> &[u8] {
+        &self.data
+    }
+
+    /// Returns a boolean indicating whether the integer is negative.
+    pub fn is_negative(&self) -> bool {
+        self.data[0] & 0x80 == 0x80
+    }
+}
+
+impl SimpleAsn1Readable<'_> for OwnedBigInt {
+    const TAG: Tag = Tag::primitive(0x02);
+    fn parse_data(data: &[u8]) -> ParseResult<Self> {
+        OwnedBigInt::new(data.to_vec()).ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
+    }
+}
+impl SimpleAsn1Writable for OwnedBigInt {
+    const TAG: Tag = Tag::primitive(0x02);
+    fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
+        dest.push_slice(self.as_bytes())
+    }
+}
+
 impl<'a> SimpleAsn1Readable<'a> for ObjectIdentifier {
     const TAG: Tag = Tag::primitive(0x06);
     fn parse_data(data: &'a [u8]) -> ParseResult<ObjectIdentifier> {
@@ -752,20 +913,86 @@ fn push_four_digits(dest: &mut WriteBuf, val: u16) -> WriteResult {
     dest.push_byte(b'0' + (val % 10) as u8)
 }
 
-/// Used for parsing and writing ASN.1 `UTC TIME` values. Wraps a
-/// `chrono::DateTime<Utc>`.
-#[derive(Debug, Clone, PartialEq, Hash, Eq)]
-pub struct UtcTime(chrono::DateTime<chrono::Utc>);
+/// A structure representing a (UTC timezone) date and time.
+/// Wrapped by `UtcTime` and `GeneralizedTime`.
+#[derive(Debug, Clone, PartialEq, Hash, Eq, PartialOrd)]
+pub struct DateTime {
+    year: u16,
+    month: u8,
+    day: u8,
+    hour: u8,
+    minute: u8,
+    second: u8,
+}
 
-impl UtcTime {
-    pub fn new(v: chrono::DateTime<chrono::Utc>) -> Option<UtcTime> {
-        if v.year() >= 2050 || v.year() < 1950 {
-            return None;
+impl DateTime {
+    pub fn new(
+        year: u16,
+        month: u8,
+        day: u8,
+        hour: u8,
+        minute: u8,
+        second: u8,
+    ) -> ParseResult<DateTime> {
+        validate_date(year, month, day)?;
+        if hour > 23 || minute > 59 || second > 59 {
+            return Err(ParseError::new(ParseErrorKind::InvalidValue));
         }
-        Some(UtcTime(v))
+        Ok(DateTime {
+            year,
+            month,
+            day,
+            hour,
+            minute,
+            second,
+        })
     }
 
-    pub fn as_chrono(&self) -> &chrono::DateTime<chrono::Utc> {
+    /// The calendar year.
+    pub fn year(&self) -> u16 {
+        self.year
+    }
+
+    /// The calendar month (1 to 12).
+    pub fn month(&self) -> u8 {
+        self.month
+    }
+
+    /// The calendar day (1 to 31).
+    pub fn day(&self) -> u8 {
+        self.day
+    }
+
+    /// The clock hour (0 to 23).
+    pub fn hour(&self) -> u8 {
+        self.hour
+    }
+
+    /// The clock minute (0 to 59).
+    pub fn minute(&self) -> u8 {
+        self.minute
+    }
+
+    /// The clock second (0 to 59).
+    pub fn second(&self) -> u8 {
+        self.second
+    }
+}
+
+/// Used for parsing and writing ASN.1 `UTC TIME` values. Wraps a
+/// `DateTime`.
+#[derive(Debug, Clone, PartialEq, Hash, Eq)]
+pub struct UtcTime(DateTime);
+
+impl UtcTime {
+    pub fn new(dt: DateTime) -> ParseResult<UtcTime> {
+        if dt.year() < 1950 || dt.year() >= 2050 {
+            return Err(ParseError::new(ParseErrorKind::InvalidValue));
+        }
+        Ok(UtcTime(dt))
+    }
+
+    pub fn as_datetime(&self) -> &DateTime {
         &self.0
     }
 }
@@ -780,63 +1007,49 @@ impl SimpleAsn1Readable<'_> for UtcTime {
         // year ordinals to full year:
         // https://tools.ietf.org/html/rfc5280#section-4.1.2.5.1
         let year = if year >= 50 { 1900 + year } else { 2000 + year };
-        validate_date(year, month, day)?;
-
         let hour = read_2_digits(&mut data)?;
         let minute = read_2_digits(&mut data)?;
         let second = read_2_digits(&mut data)?;
-        if hour > 23 || minute > 59 || second > 59 {
-            return Err(ParseError::new(ParseErrorKind::InvalidValue));
-        }
 
         read_tz_and_finish(&mut data)?;
 
-        UtcTime::new(
-            chrono::Utc
-                .ymd(year.into(), month.into(), day.into())
-                .and_hms(hour.into(), minute.into(), second.into()),
-        )
-        .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
+        UtcTime::new(DateTime::new(year, month, day, hour, minute, second)?)
     }
 }
 
 impl SimpleAsn1Writable for UtcTime {
     const TAG: Tag = Tag::primitive(0x17);
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
-        let year = if 1950 <= self.0.year() && self.0.year() < 2000 {
-            self.0.year() - 1900
+        let dt = self.as_datetime();
+        let year = if 1950 <= dt.year() && dt.year() < 2000 {
+            dt.year() - 1900
         } else {
-            assert!(2000 <= self.0.year() && self.0.year() < 2050);
-            self.0.year() - 2000
+            assert!(2000 <= dt.year() && dt.year() < 2050);
+            dt.year() - 2000
         };
         push_two_digits(dest, year.try_into().unwrap())?;
-        push_two_digits(dest, self.0.month().try_into().unwrap())?;
-        push_two_digits(dest, self.0.day().try_into().unwrap())?;
+        push_two_digits(dest, dt.month())?;
+        push_two_digits(dest, dt.day())?;
 
-        push_two_digits(dest, self.0.hour().try_into().unwrap())?;
-        push_two_digits(dest, self.0.minute().try_into().unwrap())?;
-        push_two_digits(dest, self.0.second().try_into().unwrap())?;
+        push_two_digits(dest, dt.hour())?;
+        push_two_digits(dest, dt.minute())?;
+        push_two_digits(dest, dt.second())?;
 
         dest.push_byte(b'Z')
     }
 }
 
 /// Used for parsing and writing ASN.1 `GENERALIZED TIME` values. Wraps a
-/// `chrono::DateTime<Utc>`.
+/// `DateTime`.
 #[derive(Debug, Clone, PartialEq, Hash, Eq)]
-pub struct GeneralizedTime(chrono::DateTime<chrono::Utc>);
+pub struct GeneralizedTime(DateTime);
 
 impl GeneralizedTime {
-    pub fn new(v: chrono::DateTime<chrono::Utc>) -> ParseResult<GeneralizedTime> {
-        // Reject leap seconds, which aren't allowed by ASN.1. chrono encodes
-        // them as nanoseconds == 1000000.
-        if v.year() < 0 || v.nanosecond() >= 1_000_000 {
-            return Err(ParseError::new(ParseErrorKind::InvalidValue));
-        }
-        Ok(GeneralizedTime(v))
+    pub fn new(dt: DateTime) -> ParseResult<GeneralizedTime> {
+        Ok(GeneralizedTime(dt))
     }
 
-    pub fn as_chrono(&self) -> &chrono::DateTime<chrono::Utc> {
+    pub fn as_datetime(&self) -> &DateTime {
         &self.0
     }
 }
@@ -847,36 +1060,27 @@ impl SimpleAsn1Readable<'_> for GeneralizedTime {
         let year = read_4_digits(&mut data)?;
         let month = read_2_digits(&mut data)?;
         let day = read_2_digits(&mut data)?;
-
-        validate_date(year, month, day)?;
-
         let hour = read_2_digits(&mut data)?;
         let minute = read_2_digits(&mut data)?;
         let second = read_2_digits(&mut data)?;
-        if hour > 23 || minute > 59 || second > 59 {
-            return Err(ParseError::new(ParseErrorKind::InvalidValue));
-        }
 
         read_tz_and_finish(&mut data)?;
 
-        GeneralizedTime::new(
-            chrono::Utc
-                .ymd(year.into(), month.into(), day.into())
-                .and_hms(hour.into(), minute.into(), second.into()),
-        )
+        GeneralizedTime::new(DateTime::new(year, month, day, hour, minute, second)?)
     }
 }
 
 impl SimpleAsn1Writable for GeneralizedTime {
     const TAG: Tag = Tag::primitive(0x18);
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
-        push_four_digits(dest, self.0.year().try_into().unwrap())?;
-        push_two_digits(dest, self.0.month().try_into().unwrap())?;
-        push_two_digits(dest, self.0.day().try_into().unwrap())?;
+        let dt = self.as_datetime();
+        push_four_digits(dest, dt.year())?;
+        push_two_digits(dest, dt.month())?;
+        push_two_digits(dest, dt.day())?;
 
-        push_two_digits(dest, self.0.hour().try_into().unwrap())?;
-        push_two_digits(dest, self.0.minute().try_into().unwrap())?;
-        push_two_digits(dest, self.0.second().try_into().unwrap())?;
+        push_two_digits(dest, dt.hour())?;
+        push_two_digits(dest, dt.minute())?;
+        push_two_digits(dest, dt.second())?;
 
         dest.push_byte(b'Z')
     }
@@ -928,7 +1132,7 @@ impl<'a, T: Asn1Readable<'a>> Asn1Readable<'a> for Option<T> {
 
 impl<T: Asn1Writable> Asn1Writable for Option<T> {
     #[inline]
-    fn write(&self, w: &mut Writer) -> WriteResult {
+    fn write(&self, w: &mut Writer<'_>) -> WriteResult {
         if let Some(v) = self {
             w.write_element(v)
         } else {
@@ -983,7 +1187,7 @@ macro_rules! declare_choice {
                 $number: Asn1Writable,
             )*
         > Asn1Writable for $count<$($number,)*> {
-            fn write(&self, w: &mut Writer) -> WriteResult {
+            fn write(&self, w: &mut Writer<'_>) -> WriteResult {
                 match self {
                     $(
                         $count::$name(v) => w.write_element(v),
@@ -1040,12 +1244,12 @@ impl<'a> SimpleAsn1Writable for Sequence<'a> {
 /// Writes an ASN.1 `SEQUENCE` using a callback that writes the inner
 /// elements.
 pub struct SequenceWriter<'a> {
-    f: &'a dyn Fn(&mut Writer) -> WriteResult,
+    f: &'a dyn Fn(&mut Writer<'_>) -> WriteResult,
 }
 
 impl<'a> SequenceWriter<'a> {
     #[inline]
-    pub fn new(f: &'a dyn Fn(&mut Writer) -> WriteResult) -> Self {
+    pub fn new(f: &'a dyn Fn(&mut Writer<'_>) -> WriteResult) -> Self {
         SequenceWriter { f }
     }
 }
@@ -1060,15 +1264,22 @@ impl<'a> SimpleAsn1Writable for SequenceWriter<'a> {
 
 /// Represents an ASN.1 `SEQUENCE OF`. This is an `Iterator` over values that
 /// are decoded.
-pub struct SequenceOf<'a, T: Asn1Readable<'a>> {
+pub struct SequenceOf<
+    'a,
+    T: Asn1Readable<'a>,
+    const MINIMUM_LEN: usize = 0,
+    const MAXIMUM_LEN: usize = { usize::MAX },
+> {
     parser: Parser<'a>,
     length: usize,
     _phantom: PhantomData<T>,
 }
 
-impl<'a, T: Asn1Readable<'a>> SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>, const MINIMUM_LEN: usize, const MAXIMUM_LEN: usize>
+    SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
     #[inline]
-    pub(crate) fn new(data: &'a [u8]) -> ParseResult<SequenceOf<'a, T>> {
+    pub(crate) fn new(data: &'a [u8]) -> ParseResult<SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>> {
         let length = parse(data, |p| {
             let mut i = 0;
             while !p.is_empty() {
@@ -1079,7 +1290,15 @@ impl<'a, T: Asn1Readable<'a>> SequenceOf<'a, T> {
             Ok(i)
         })?;
 
-        Ok(SequenceOf {
+        if length < MINIMUM_LEN || length > MAXIMUM_LEN {
+            return Err(ParseError::new(ParseErrorKind::InvalidSize {
+                min: MINIMUM_LEN,
+                max: MAXIMUM_LEN,
+                actual: length,
+            }));
+        }
+
+        Ok(Self {
             length,
             parser: Parser::new(data),
             _phantom: PhantomData,
@@ -1095,8 +1314,10 @@ impl<'a, T: Asn1Readable<'a>> SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: Asn1Readable<'a>> Clone for SequenceOf<'a, T> {
-    fn clone(&self) -> SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>, const MINIMUM_LEN: usize, const MAXIMUM_LEN: usize> Clone
+    for SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
+    fn clone(&self) -> SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN> {
         SequenceOf {
             parser: self.parser.clone_internal(),
             length: self.length,
@@ -1105,7 +1326,9 @@ impl<'a, T: Asn1Readable<'a>> Clone for SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: Asn1Readable<'a> + PartialEq> PartialEq for SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a> + PartialEq, const MINIMUM_LEN: usize, const MAXIMUM_LEN: usize>
+    PartialEq for SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
     fn eq(&self, other: &Self) -> bool {
         let mut it1 = self.clone();
         let mut it2 = other.clone();
@@ -1123,7 +1346,14 @@ impl<'a, T: Asn1Readable<'a> + PartialEq> PartialEq for SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: Asn1Readable<'a> + Hash> Hash for SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a> + Eq, const MINIMUM_LEN: usize, const MAXIMUM_LEN: usize> Eq
+    for SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
+}
+
+impl<'a, T: Asn1Readable<'a> + Hash, const MINIMUM_LEN: usize, const MAXIMUM_LEN: usize> Hash
+    for SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
     fn hash<H: Hasher>(&self, state: &mut H) {
         for val in self.clone() {
             val.hash(state);
@@ -1131,7 +1361,9 @@ impl<'a, T: Asn1Readable<'a> + Hash> Hash for SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a> + 'a, const MINIMUM_LEN: usize, const MAXIMUM_LEN: usize>
+    SimpleAsn1Readable<'a> for SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
     const TAG: Tag = Tag::constructed(0x10);
     #[inline]
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
@@ -1139,7 +1371,9 @@ impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SequenceOf<'a, T>
     }
 }
 
-impl<'a, T: Asn1Readable<'a>> Iterator for SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>, const MINIMUM_LEN: usize, const MAXIMUM_LEN: usize> Iterator
+    for SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
     type Item = T;
 
     fn next(&mut self) -> Option<Self::Item> {
@@ -1155,7 +1389,13 @@ impl<'a, T: Asn1Readable<'a>> Iterator for SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: Asn1Readable<'a> + Asn1Writable> SimpleAsn1Writable for SequenceOf<'a, T> {
+impl<
+        'a,
+        T: Asn1Readable<'a> + Asn1Writable,
+        const MINIMUM_LEN: usize,
+        const MAXIMUM_LEN: usize,
+    > SimpleAsn1Writable for SequenceOf<'a, T, MINIMUM_LEN, MAXIMUM_LEN>
+{
     const TAG: Tag = Tag::constructed(0x10);
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         let mut w = Writer::new(dest);
@@ -1239,6 +1479,8 @@ impl<'a, T: Asn1Readable<'a> + PartialEq> PartialEq for SetOf<'a, T> {
     }
 }
 
+impl<'a, T: Asn1Readable<'a> + Eq> Eq for SetOf<'a, T> {}
+
 impl<'a, T: Asn1Readable<'a> + Hash> Hash for SetOf<'a, T> {
     fn hash<H: Hasher>(&self, state: &mut H) {
         for val in self.clone() {
@@ -1253,7 +1495,7 @@ impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SetOf<'a, T> {
     #[inline]
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
         parse(data, |p| {
-            let mut last_element: Option<Tlv> = None;
+            let mut last_element: Option<Tlv<'a>> = None;
             let mut i = 0;
             while !p.is_empty() {
                 let el = p
@@ -1359,41 +1601,33 @@ impl<'a, T: Asn1Writable, V: Borrow<[T]>> SimpleAsn1Writable for SetOfWriter<'a,
 
 /// `Implicit` is a type which wraps another ASN.1 type, indicating that the tag is an ASN.1
 /// `IMPLICIT`. This will generally be used with `Option` or `Choice`.
-///
-/// Requires the `const-generics` feature and Rust 1.51 or greater. For users
-/// on older Rust versions, `Parser::read_optional_implicit_element` may be
-/// used.
-#[cfg(feature = "const-generics")]
 #[derive(PartialEq, Eq, Debug)]
-pub struct Implicit<'a, T, const TAG: u32> {
+pub struct Implicit<T, const TAG: u32> {
     inner: T,
-    _lifetime: PhantomData<&'a ()>,
 }
 
-#[cfg(feature = "const-generics")]
-impl<'a, T, const TAG: u32> Implicit<'a, T, { TAG }> {
+impl<T, const TAG: u32> Implicit<T, { TAG }> {
     pub fn new(v: T) -> Self {
-        Implicit {
-            inner: v,
-            _lifetime: PhantomData,
-        }
+        Implicit { inner: v }
     }
 
     pub fn as_inner(&self) -> &T {
         &self.inner
     }
+
+    pub fn into_inner(self) -> T {
+        self.inner
+    }
 }
 
-#[cfg(feature = "const-generics")]
-impl<'a, T, const TAG: u32> From<T> for Implicit<'a, T, { TAG }> {
+impl<T, const TAG: u32> From<T> for Implicit<T, { TAG }> {
     fn from(v: T) -> Self {
         Implicit::new(v)
     }
 }
 
-#[cfg(feature = "const-generics")]
 impl<'a, T: SimpleAsn1Readable<'a>, const TAG: u32> SimpleAsn1Readable<'a>
-    for Implicit<'a, T, { TAG }>
+    for Implicit<T, { TAG }>
 {
     const TAG: Tag = crate::implicit_tag(TAG, T::TAG);
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
@@ -1401,8 +1635,7 @@ impl<'a, T: SimpleAsn1Readable<'a>, const TAG: u32> SimpleAsn1Readable<'a>
     }
 }
 
-#[cfg(feature = "const-generics")]
-impl<'a, T: SimpleAsn1Writable, const TAG: u32> SimpleAsn1Writable for Implicit<'a, T, { TAG }> {
+impl<T: SimpleAsn1Writable, const TAG: u32> SimpleAsn1Writable for Implicit<T, { TAG }> {
     const TAG: Tag = crate::implicit_tag(TAG, T::TAG);
 
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
@@ -1412,66 +1645,101 @@ impl<'a, T: SimpleAsn1Writable, const TAG: u32> SimpleAsn1Writable for Implicit<
 
 /// `Explicit` is a type which wraps another ASN.1 type, indicating that the tag is an ASN.1
 /// `EXPLICIT`. This will generally be used with `Option` or `Choice`.
-///
-/// Requires the `const-generics` feature and Rust 1.51 or greater. For users
-/// on older Rust versions, `Parser::read_optional_explicit_element` may be
-/// used.
-#[cfg(feature = "const-generics")]
 #[derive(PartialEq, Eq, Debug)]
-pub struct Explicit<'a, T, const TAG: u32> {
+pub struct Explicit<T, const TAG: u32> {
     inner: T,
-    _lifetime: PhantomData<&'a ()>,
 }
 
-#[cfg(feature = "const-generics")]
-impl<'a, T, const TAG: u32> Explicit<'a, T, { TAG }> {
+impl<T, const TAG: u32> Explicit<T, { TAG }> {
     pub fn new(v: T) -> Self {
-        Explicit {
-            inner: v,
-            _lifetime: PhantomData,
-        }
+        Explicit { inner: v }
     }
 
     pub fn as_inner(&self) -> &T {
         &self.inner
     }
+
+    pub fn into_inner(self) -> T {
+        self.inner
+    }
 }
 
-#[cfg(feature = "const-generics")]
-impl<'a, T, const TAG: u32> From<T> for Explicit<'a, T, { TAG }> {
+impl<T, const TAG: u32> From<T> for Explicit<T, { TAG }> {
     fn from(v: T) -> Self {
         Explicit::new(v)
     }
 }
 
-#[cfg(feature = "const-generics")]
-impl<'a, T: Asn1Readable<'a>, const TAG: u32> SimpleAsn1Readable<'a> for Explicit<'a, T, { TAG }> {
+impl<'a, T: Asn1Readable<'a>, const TAG: u32> SimpleAsn1Readable<'a> for Explicit<T, { TAG }> {
     const TAG: Tag = crate::explicit_tag(TAG);
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
-        Ok(Explicit::new(parse(data, |p| p.read_element::<T>())?))
+        Ok(Explicit::new(parse(data, Parser::read_element::<T>)?))
     }
 }
 
-#[cfg(feature = "const-generics")]
-impl<'a, T: Asn1Writable, const TAG: u32> SimpleAsn1Writable for Explicit<'a, T, { TAG }> {
+impl<T: Asn1Writable, const TAG: u32> SimpleAsn1Writable for Explicit<T, { TAG }> {
     const TAG: Tag = crate::explicit_tag(TAG);
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         Writer::new(dest).write_element(&self.inner)
     }
 }
 
+impl<'a, T: Asn1Readable<'a>, U: Asn1DefinedByReadable<'a, T>, const TAG: u32>
+    Asn1DefinedByReadable<'a, T> for Explicit<U, { TAG }>
+{
+    fn parse(item: T, parser: &mut Parser<'a>) -> ParseResult<Self> {
+        let tlv = parser.read_element::<Explicit<Tlv<'_>, TAG>>()?;
+        Ok(Explicit::new(parse(tlv.as_inner().full_data(), |p| {
+            U::parse(item, p)
+        })?))
+    }
+}
+
+impl<T: Asn1Writable, U: Asn1DefinedByWritable<T>, const TAG: u32> Asn1DefinedByWritable<T>
+    for Explicit<U, { TAG }>
+{
+    fn item(&self) -> &T {
+        self.as_inner().item()
+    }
+    fn write(&self, dest: &mut Writer<'_>) -> WriteResult {
+        dest.write_tlv(crate::explicit_tag(TAG), |dest| {
+            self.as_inner().write(&mut Writer::new(dest))
+        })
+    }
+}
+
+#[derive(PartialEq, Eq, Debug, Clone, Hash)]
+pub struct DefinedByMarker<T>(core::marker::PhantomData<T>);
+
+impl<T> DefinedByMarker<T> {
+    pub const fn marker() -> DefinedByMarker<T> {
+        DefinedByMarker(core::marker::PhantomData)
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use crate::{
-        parse_single, BigInt, BigUint, Enumerated, GeneralizedTime, IA5String, ParseError,
+        parse_single, BigInt, BigUint, DateTime, DefinedByMarker, Enumerated, GeneralizedTime,
+        IA5String, ObjectIdentifier, OctetStringEncoded, OwnedBigInt, OwnedBigUint, ParseError,
         ParseErrorKind, PrintableString, SequenceOf, SequenceOfWriter, SetOf, SetOfWriter, Tag,
         Tlv, UtcTime, Utf8String, VisibleString,
     };
-    #[cfg(feature = "const-generics")]
     use crate::{Explicit, Implicit};
-    use chrono::{TimeZone, Utc};
+    #[cfg(not(feature = "std"))]
+    use alloc::vec;
+    #[cfg(not(feature = "std"))]
+    use alloc::vec::Vec;
+    #[cfg(feature = "std")]
+    use core::hash::{Hash, Hasher};
+    #[cfg(feature = "std")]
     use std::collections::hash_map::DefaultHasher;
-    use std::hash::{Hash, Hasher};
+
+    #[test]
+    fn test_octet_string_encoded() {
+        assert_eq!(OctetStringEncoded::new(12).get(), &12);
+        assert_eq!(OctetStringEncoded::new(12).into_inner(), 12);
+    }
 
     #[test]
     fn test_printable_string_new() {
@@ -1518,6 +1786,12 @@ mod tests {
         assert_eq!(VisibleString::new("abc").unwrap().as_str(), "abc");
     }
 
+    #[test]
+    fn test_tlv_data() {
+        let tlv = parse_single::<Tlv<'_>>(b"\x01\x03abc").unwrap();
+        assert_eq!(tlv.data(), b"abc");
+    }
+
     #[test]
     fn test_tlv_full_data() {
         let tlv = parse_single::<Tlv<'_>>(b"\x01\x03abc").unwrap();
@@ -1543,17 +1817,36 @@ mod tests {
     #[test]
     fn test_biguint_as_bytes() {
         assert_eq!(BigUint::new(b"\x01").unwrap().as_bytes(), b"\x01");
+        assert_eq!(
+            OwnedBigUint::new(b"\x01".to_vec()).unwrap().as_bytes(),
+            b"\x01"
+        );
     }
 
     #[test]
     fn test_bigint_as_bytes() {
         assert_eq!(BigInt::new(b"\x01").unwrap().as_bytes(), b"\x01");
+        assert_eq!(
+            OwnedBigInt::new(b"\x01".to_vec()).unwrap().as_bytes(),
+            b"\x01"
+        );
+    }
+
+    #[test]
+    fn test_bigint_is_negative() {
+        assert!(!BigInt::new(b"\x01").unwrap().is_negative()); // 1
+        assert!(!BigInt::new(b"\x00").unwrap().is_negative()); // 0
+        assert!(BigInt::new(b"\xff").unwrap().is_negative()); // -1
+
+        assert!(!OwnedBigInt::new(b"\x01".to_vec()).unwrap().is_negative()); // 1
+        assert!(!OwnedBigInt::new(b"\x00".to_vec()).unwrap().is_negative()); // 0
+        assert!(OwnedBigInt::new(b"\xff".to_vec()).unwrap().is_negative()); // -1
     }
 
     #[test]
     fn test_sequence_of_clone() {
         let mut seq1 =
-            parse_single::<SequenceOf<u64>>(b"\x30\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03")
+            parse_single::<SequenceOf<'_, u64>>(b"\x30\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03")
                 .unwrap();
         assert_eq!(seq1.next(), Some(1));
         let seq2 = seq1.clone();
@@ -1564,7 +1857,7 @@ mod tests {
     #[test]
     fn test_sequence_of_len() {
         let mut seq1 =
-            parse_single::<SequenceOf<u64>>(b"\x30\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03")
+            parse_single::<SequenceOf<'_, u64>>(b"\x30\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03")
                 .unwrap();
         let seq2 = seq1.clone();
 
@@ -1582,6 +1875,7 @@ mod tests {
         assert!(!seq2.is_empty());
     }
 
+    #[cfg(feature = "std")]
     fn hash<T: Hash>(v: &T) -> u64 {
         let mut h = DefaultHasher::new();
         v.hash(&mut h);
@@ -1589,40 +1883,64 @@ mod tests {
     }
 
     #[test]
-    fn test_set_of_eq_hash() {
+    fn test_set_of_eq() {
         let s1 = SetOf::<bool>::new(b"");
         let s2 = SetOf::<bool>::new(b"");
         let s3 = SetOf::<bool>::new(b"\x01\x01\x00");
         let s4 = SetOf::<bool>::new(b"\x01\x01\xff");
 
         assert!(s1 == s2);
-        assert_eq!(hash(&s1), hash(&s2));
 
         assert!(s2 != s3);
-        assert_ne!(hash(&s2), hash(&s3));
 
         assert!(s3 == s3);
 
         assert!(s3 != s4);
+    }
+
+    #[cfg(feature = "std")]
+    #[test]
+    fn test_set_of_hash() {
+        let s1 = SetOf::<bool>::new(b"");
+        let s2 = SetOf::<bool>::new(b"");
+        let s3 = SetOf::<bool>::new(b"\x01\x01\x00");
+        let s4 = SetOf::<bool>::new(b"\x01\x01\xff");
+
+        assert_eq!(hash(&s1), hash(&s2));
+
+        assert_ne!(hash(&s2), hash(&s3));
+
         assert_ne!(hash(&s3), hash(&s4));
     }
 
     #[test]
-    fn test_sequence_of_eq_hash() {
+    fn test_sequence_of_eq() {
         let s1 = SequenceOf::<bool>::new(b"").unwrap();
         let s2 = SequenceOf::<bool>::new(b"").unwrap();
         let s3 = SequenceOf::<bool>::new(b"\x01\x01\x00").unwrap();
         let s4 = SequenceOf::<bool>::new(b"\x01\x01\xff").unwrap();
 
         assert!(s1 == s2);
-        assert_eq!(hash(&s1), hash(&s2));
 
         assert!(s2 != s3);
-        assert_ne!(hash(&s2), hash(&s3));
 
         assert!(s3 == s3);
 
         assert!(s3 != s4);
+    }
+
+    #[cfg(feature = "std")]
+    #[test]
+    fn test_sequence_of_hash() {
+        let s1 = SequenceOf::<bool>::new(b"").unwrap();
+        let s2 = SequenceOf::<bool>::new(b"").unwrap();
+        let s3 = SequenceOf::<bool>::new(b"\x01\x01\x00").unwrap();
+        let s4 = SequenceOf::<bool>::new(b"\x01\x01\xff").unwrap();
+
+        assert_eq!(hash(&s1), hash(&s2));
+
+        assert_ne!(hash(&s2), hash(&s3));
+
         assert_ne!(hash(&s3), hash(&s4));
     }
 
@@ -1641,30 +1959,45 @@ mod tests {
 
         assert!(s1 == s2);
     }
+
     #[test]
-    fn test_utctime_new() {
-        assert!(UtcTime::new(chrono::Utc.ymd(1950, 1, 1).and_hms(12, 0, 0)).is_some());
-        assert!(UtcTime::new(chrono::Utc.ymd(2050, 1, 1).and_hms(12, 0, 0)).is_none());
+    fn test_datetime_new() {
+        assert!(DateTime::new(2038, 13, 1, 12, 0, 0).is_err());
+        assert!(DateTime::new(2000, 1, 1, 12, 60, 0).is_err());
+        assert!(DateTime::new(2000, 1, 1, 12, 0, 60).is_err());
+        assert!(DateTime::new(2000, 1, 1, 24, 0, 0).is_err());
     }
 
     #[test]
-    fn test_utctime_as_chrono() {
-        let t = Utc.ymd(1951, 5, 6).and_hms(23, 45, 0);
-        assert_eq!(UtcTime::new(t).unwrap().as_chrono(), &t);
+    fn test_datetime_partialord() {
+        let point = DateTime::new(2023, 6, 15, 14, 26, 5).unwrap();
+
+        assert!(point < DateTime::new(2023, 6, 15, 14, 26, 6).unwrap());
+        assert!(point < DateTime::new(2023, 6, 15, 14, 27, 5).unwrap());
+        assert!(point < DateTime::new(2023, 6, 15, 15, 26, 5).unwrap());
+        assert!(point < DateTime::new(2023, 6, 16, 14, 26, 5).unwrap());
+        assert!(point < DateTime::new(2023, 7, 15, 14, 26, 5).unwrap());
+        assert!(point < DateTime::new(2024, 6, 15, 14, 26, 5).unwrap());
+
+        assert!(point > DateTime::new(2023, 6, 15, 14, 26, 4).unwrap());
+        assert!(point > DateTime::new(2023, 6, 15, 14, 25, 5).unwrap());
+        assert!(point > DateTime::new(2023, 6, 15, 13, 26, 5).unwrap());
+        assert!(point > DateTime::new(2023, 6, 14, 14, 26, 5).unwrap());
+        assert!(point > DateTime::new(2023, 5, 15, 14, 26, 5).unwrap());
+        assert!(point > DateTime::new(2022, 6, 15, 14, 26, 5).unwrap());
+    }
+
+    #[test]
+    fn test_utctime_new() {
+        assert!(UtcTime::new(DateTime::new(1950, 1, 1, 12, 0, 0).unwrap()).is_ok());
+        assert!(UtcTime::new(DateTime::new(1949, 1, 1, 12, 0, 0).unwrap()).is_err());
+        assert!(UtcTime::new(DateTime::new(2050, 1, 1, 12, 0, 0).unwrap()).is_err());
+        assert!(UtcTime::new(DateTime::new(2100, 1, 1, 12, 0, 0).unwrap()).is_err());
     }
 
     #[test]
     fn test_generalized_time_new() {
-        let t = Utc.ymd(2015, 6, 30).and_hms_nano(23, 59, 59, 1_000_000_000);
-        assert!(GeneralizedTime::new(t).is_err());
-        let t = Utc.ymd(2015, 6, 30).and_hms_nano(23, 59, 59, 0);
-        assert!(GeneralizedTime::new(t).is_ok());
-    }
-
-    #[test]
-    fn test_generalized_time_as_chrono() {
-        let t = Utc.ymd(1951, 5, 6).and_hms(23, 45, 0);
-        assert_eq!(GeneralizedTime::new(t).unwrap().as_chrono(), &t);
+        assert!(GeneralizedTime::new(DateTime::new(2015, 6, 30, 23, 59, 59).unwrap()).is_ok());
     }
 
     #[test]
@@ -1673,14 +2006,17 @@ mod tests {
     }
 
     #[test]
-    #[cfg(feature = "const-generics")]
     fn test_implicit_as_inner() {
         assert_eq!(Implicit::<i32, 0>::new(12).as_inner(), &12);
     }
 
     #[test]
-    #[cfg(feature = "const-generics")]
     fn test_explicit_as_inner() {
         assert_eq!(Explicit::<i32, 0>::new(12).as_inner(), &12);
     }
+
+    #[test]
+    fn test_const() {
+        const _: DefinedByMarker<ObjectIdentifier> = DefinedByMarker::marker();
+    }
 }
